"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Protocol = exports.STATE_SEED = void 0;
const web3_js_1 = require("@solana/web3.js");
const anchor_1 = require("@project-serum/anchor");
const fabric_synths_json_1 = __importDefault(require("./idl/fabric_synths.json"));
const spl_token_1 = require("@solana/spl-token");
// Constants
exports.STATE_SEED = "fstate1";
// 
class Protocol {
    constructor(connection, wallet, programId, opts) {
        this.idl = fabric_synths_json_1.default;
        this.wallet = wallet;
        this.programId = programId;
        this.connection = connection;
        const provider = new anchor_1.Provider(this.connection, this.wallet, opts || anchor_1.Provider.defaultOptions());
        this.program = new anchor_1.Program(fabric_synths_json_1.default, this.programId, provider);
    }
    static initialise(connection, wallet, programId, opts) {
        return __awaiter(this, void 0, void 0, function* () {
            var protocol = new Protocol(connection, wallet, programId, opts);
            protocol.statePubkey = (yield protocol.getStateAccount())[0];
            return protocol;
        });
    }
    // ix methods
    initialiseContract(nonce) {
        return __awaiter(this, void 0, void 0, function* () {
            var stateAddressInfo = yield this.getStateAccount();
            this.adminPubkey = this.wallet.publicKey;
            var tx = yield this.program.rpc.initialiseContract(new anchor_1.BN(stateAddressInfo[1]), new anchor_1.BN(nonce), {
                accounts: {
                    stateAccount: stateAddressInfo[0],
                    adminAccount: this.adminPubkey,
                    systemProgram: web3_js_1.SystemProgram.programId
                }
            });
            this.statePubkey = stateAddressInfo[0];
            return tx;
        });
    }
    initialiseAssetsList(collateralMintPubkey, collateralFeedPubkey, collateralReserveAccountPubkey, collateralLiquidationFundAccountPubkey) {
        return __awaiter(this, void 0, void 0, function* () {
            const assetListAccount = web3_js_1.Keypair.generate();
            var stateAddressInfo = yield this.getStateAccount();
            let usdToken = yield this.createToken(8);
            let tx = yield this.program.rpc.initialiseAssetsList({
                accounts: {
                    stateAccount: stateAddressInfo[0],
                    assetsList: assetListAccount.publicKey,
                    adminAccount: this.wallet.publicKey,
                    collateralToken: collateralMintPubkey,
                    collateralTokenFeed: collateralFeedPubkey,
                    usdToken: usdToken.publicKey,
                    fabReserve: collateralReserveAccountPubkey,
                    fabLiquidationFund: collateralLiquidationFundAccountPubkey,
                    exchangeAuthority: this.wallet.publicKey,
                    rent: web3_js_1.SYSVAR_RENT_PUBKEY
                },
                signers: [
                    assetListAccount,
                    //  @ts-expect-error
                    this.program.provider.wallet.payer,
                ],
                instructions: [
                    yield this.program.account.assetsList.createInstruction(assetListAccount)
                ]
            });
            this.assetsListPubkey = assetListAccount.publicKey;
            let result = [
                tx,
                usdToken,
                assetListAccount.publicKey
            ];
            return result;
        });
    }
    setAssetsListAccount(assetListPubkey) {
        return __awaiter(this, void 0, void 0, function* () {
            let tx = yield this.program.rpc.setAssetsList({
                accounts: {
                    stateAccount: this.statePubkey,
                    adminAccount: this.adminPubkey,
                    assetsList: assetListPubkey
                }
            });
            this.assetsListPubkey = assetListPubkey;
            return tx;
        });
    }
    depositCollateral(amount, userAccount, collateralReserveAccount, userCollateralAccount, owner) {
        return __awaiter(this, void 0, void 0, function* () {
            let tx = yield this.program.rpc.deposit(amount, {
                accounts: {
                    stateAccount: this.statePubkey,
                    userAccount: userAccount,
                    collateralReserveAccount,
                    userCollateralAccount,
                    tokenProgramAccount: spl_token_1.TOKEN_PROGRAM_ID,
                    assetsListAccount: this.assetsListPubkey,
                    owner,
                    adminAccount: this.adminPubkey,
                }
            });
            return tx;
        });
    }
    mint(amount, owner, fusdReceiver, userAccount) {
        return __awaiter(this, void 0, void 0, function* () {
            let tx = yield this.program.rpc.mint(amount, {
                accounts: {
                    stateAccount: this.statePubkey,
                    adminAccount: this.adminPubkey,
                    assetsListAccount: this.assetsListPubkey,
                    fusdTokenAccount: (yield this.getfUsdTokenMint()),
                    toAccount: fusdReceiver,
                    userAccount: userAccount,
                    tokenProgramAccount: spl_token_1.TOKEN_PROGRAM_ID,
                    owner,
                }
            });
            return tx;
        });
    }
    burn(amount, owner, fusdReceiver, userAccount) {
        return __awaiter(this, void 0, void 0, function* () {
            var tx = yield this.program.rpc.burn(amount, {
                accounts: {
                    stateAccount: this.statePubkey,
                    assetsListAccount: this.assetsListPubkey,
                    adminAccount: this.adminPubkey,
                    fusdTokenAccount: (yield this.getfUsdTokenMint()),
                    fusdToBurnAccount: fusdReceiver,
                    tokenProgramAccount: spl_token_1.TOKEN_PROGRAM_ID,
                    userAccount: userAccount,
                    owner: owner,
                }
            });
            return tx;
        });
    }
    withdrawCollateral(amount, userAccount, collateralReserveAccount, userCollateralAccount, owner) {
        return __awaiter(this, void 0, void 0, function* () {
            var tx = yield this.program.rpc.withdraw(amount, {
                accounts: {
                    stateAccount: this.statePubkey,
                    userAccount: userAccount,
                    collateralReserveAccount,
                    userCollateralAccount,
                    tokenProgramAccount: spl_token_1.TOKEN_PROGRAM_ID,
                    assetsListAccount: this.assetsListPubkey,
                    owner,
                    adminAccount: this.adminPubkey,
                }
            });
            return tx;
        });
    }
    swap(amount, owner, tokenAMint, tokenBMint, tokenAReceiver, tokenBReceiver, userAccount) {
        return __awaiter(this, void 0, void 0, function* () {
            var tx = yield this.program.rpc.swap(amount, {
                accounts: {
                    stateAccount: this.statePubkey,
                    adminAccount: this.adminPubkey,
                    assetsListAccount: this.assetsListPubkey,
                    owner,
                    tokenProgramAccount: spl_token_1.TOKEN_PROGRAM_ID,
                    tokenInAccount: tokenAMint,
                    tokenOutAccount: tokenBMint,
                    userTokenInAccount: tokenAReceiver,
                    userTokenOutAccount: tokenBReceiver,
                    userAccount
                }
            });
            return tx;
        });
    }
    // helper methods
    getStateAccount() {
        return __awaiter(this, void 0, void 0, function* () {
            const [stateAddress, stateBump] = yield web3_js_1.PublicKey.findProgramAddress([Buffer.from(anchor_1.utils.bytes.utf8.encode(exports.STATE_SEED))], this.program.programId);
            var state = [
                stateAddress,
                stateBump
            ];
            return state;
        });
    }
    ;
    getState() {
        return __awaiter(this, void 0, void 0, function* () {
            const state = yield this.program.account.state.fetch(this.statePubkey);
            this.adminPubkey = state.adminPubkey;
            return state;
        });
    }
    getAssetList() {
        return __awaiter(this, void 0, void 0, function* () {
            const state = yield this.getState();
            const assetsList = yield this.program.account.assetsList.fetch(state.assetsList);
            assetsList.assets = assetsList.assets.slice(0, assetsList.numberOfAssets);
            assetsList.collaterals = assetsList.collaterals.slice(0, assetsList.numberOfCollaterals);
            assetsList.synthetics = assetsList.synthetics.slice(0, assetsList.numberOfSynthetics);
            return assetsList;
        });
    }
    createUserAccount(owner) {
        return __awaiter(this, void 0, void 0, function* () {
            var [userAddress, userBump] = yield this.getUserAccount(owner);
            var tx = yield this.program.rpc.createUserAccount(userBump, {
                accounts: {
                    userAccount: userAddress,
                    adminAccount: owner,
                    payer: this.wallet.publicKey,
                    rent: web3_js_1.SYSVAR_RENT_PUBKEY,
                    systemProgram: web3_js_1.SystemProgram.programId,
                }
            });
            yield this.program.provider.connection.confirmTransaction(tx);
            return userAddress;
        });
    }
    getUserAccount(owner) {
        return __awaiter(this, void 0, void 0, function* () {
            const [account, bump] = yield web3_js_1.PublicKey.findProgramAddress([Buffer.from(anchor_1.utils.bytes.utf8.encode('faccv1')), owner.toBuffer()], this.program.programId);
            let result = [
                account,
                bump
            ];
            return result;
        });
    }
    getUser(userAccount) {
        return __awaiter(this, void 0, void 0, function* () {
            const user = yield this.program.account.userAccount.fetch(userAccount);
            return user;
        });
    }
    checkIfUserAccountExists(owner) {
        return __awaiter(this, void 0, void 0, function* () {
            const userAddress = yield this.getUserAccount(owner);
            try {
                yield this.getUser(userAddress[0]);
                return true;
            }
            catch (_a) {
                return false;
            }
        });
    }
    mintTokens(amount, decimals, toAddress, token) {
        return __awaiter(this, void 0, void 0, function* () {
            // you must be the correct authority to mint tokens
            var amountToMint = new anchor_1.BN(amount * Math.pow(10, decimals));
            yield token.mintTo(toAddress, this.wallet.publicKey, [], new spl_token_1.u64(amountToMint.toString()));
        });
    }
    createTokenAccount(token, owner) {
        return __awaiter(this, void 0, void 0, function* () {
            var address = yield token.createAccount(owner);
            return address;
        });
    }
    getfUsdTokenMint() {
        return __awaiter(this, void 0, void 0, function* () {
            var assetList = yield this.getAssetList();
            return assetList.synthetics[0].assetAddress;
        });
    }
    createToken(decimals) {
        return __awaiter(this, void 0, void 0, function* () {
            const token = yield spl_token_1.Token.createMint(this.connection, 
            // @ts-expect-error
            this.program.provider.wallet.payer, this.wallet.publicKey, null, decimals, spl_token_1.TOKEN_PROGRAM_ID);
            return token;
        });
    }
    getTokenAccountInfo(token, tokenAccount) {
        return __awaiter(this, void 0, void 0, function* () {
            const info = yield token.getAccountInfo(tokenAccount);
            return info;
        });
    }
    getUserCollateralRatio(userAccount) {
        return __awaiter(this, void 0, void 0, function* () {
            const userDebtUsd = yield this.getUserDebtUsd(userAccount);
            const userCollateralUsd = yield this.getUserCollateralUsd(userAccount);
            const userCollateralRatio = (userCollateralUsd / userDebtUsd);
            return userCollateralRatio;
        });
    }
    getUserDebtUsd(userAccount) {
        return __awaiter(this, void 0, void 0, function* () {
            const user = yield this.getUser(userAccount);
            const state = yield this.getState();
            const globalDebtUsd = (yield this.getGlobalDebtUsd()).toNumber() / (Math.pow(10, 8));
            const userDebtShares = (user.userDebtShares).toNumber();
            const globalDebtShares = (state.debtShares).toNumber();
            if (globalDebtShares === 0) {
                return 0;
            }
            return (globalDebtUsd * userDebtShares / globalDebtShares);
        });
    }
    getUserCollateralUsd(userAccount) {
        return __awaiter(this, void 0, void 0, function* () {
            const user = yield this.getUser(userAccount);
            const assetList = yield this.getAssetList();
            const assets = assetList.assets;
            const collaterals = assetList.collaterals;
            const userCollaterals = user.userCollaterals;
            var sum = 0;
            userCollaterals.forEach(userCollateral => {
                const collateral = collaterals[userCollateral.index];
                const asset = assets[collateral.assetIndex];
                const amountStaked = userCollateral.amount.toNumber() * (Math.pow(10, asset.price.scale));
                const assetPrice = asset.price.val.toNumber() * (Math.pow(10, asset.price.scale));
                const amountStakedUsd = (amountStaked * assetPrice);
                sum += amountStakedUsd;
            });
            return sum;
        });
    }
    getGlobalDebtUsd() {
        return __awaiter(this, void 0, void 0, function* () {
            const assetList = yield this.getAssetList();
            var sum = new anchor_1.BN(0);
            assetList.synthetics.forEach(element => {
                const asset = assetList.assets[element.assetIndex];
                sum = sum.add(element.supply.val
                    .mul(asset.price.val)
                    .div(new anchor_1.BN(Math.pow(10, 8))));
            });
            return sum;
        });
    }
}
exports.Protocol = Protocol;
;
;
